#!/usr/bin/env python2
from __future__ import print_function

"""filedrop.py: http upload target"""
__version__   = "1.3-beta"
__author__    = "ed <filedrop@ocv.me>"
__url__       = "https://ocv.me/dev/?filedrop.py"
__credits__   = ["stackoverflow.com"]
__license__   = "MIT"
__copyright__ = 2019

##########################################
## THIS IS BUGGY, WAIT FOR THE NON-BETA ##
##########################################
#
# expect about 400 MiB/s, with py2 being slightly faster than py3
#
# navigate to your ip:port combination with a browser to see a
# basic upload form, or just use curl to upload windows7.iso:
#   curl -F "f=@windows7.iso" http://127.0.0.1:1234/
#
# the http response is a minimal html which contains the
# filesize and sha512 checksum of the uploaded file, as seen by us
#
# for a smaller alternative, see https://ocv.me/dev/?filesink.py
# which uses the built-in http parser (cgi.FieldStorage) and thus
# generates tempfiles and maxes out at 32 MiB/s (this one does 2300)
#
#-----------------------------------------------------------------------
#
# 2019-0x-xx, xx:xx, v1.3
#  + allow saturation of multiple cores through multiprocessing
#
# 2019-05-20, 00:09, v1.2
#  + allow multiple parallel transfers, see max_threads
#  + change from sha1 to truncated sha512 (slowdown to 76%)
#  + display primary local ip (default route) at startup
#
# 2019-05-13, 00:27, v1.1
#  + support multiple files
#  @ slowdown from 546 to 514 MiB/s, 94%
#
# 2019-05-01, 17:29, v1.0
#  + initial working release

import os
import sys
import time
import hashlib
import select
import socket
import signal
import threading
import multiprocessing as mp
from datetime import datetime
from multiprocessing.reduction import ForkingPickler
import pickle

##
## CONFIG  (last definition wins)

bind_port = 1234 # argv[1] overrides this

bind_ip = '127.0.0.1' # only allow local connections
bind_ip = '0.0.0.0'   # allow connections from all interfaces

nullwrite = False # regular operation
nullwrite = True  # disable all file output, for benchmark purposes

cores = 8 # allow saturation of 2 CPU cores, in addition to light work on main thread
cores = 0 # autodetect the number of logical cores; optimal performance

max_threads = 6   # max parallel transfers on each cpu core
max_threads = 128 # max parallel transfers on each cpu core

## END CONFIG
##


PY2 = (sys.version_info[0] == 2)
if PY2:
	from StringIO import StringIO as MemesIO
else:
	from io import BytesIO as MemesIO
	unicode = str


##
## DEBUG HELPERS

from inspect import currentframe

HEX_WIDTH = 32

def _lno():
	return currentframe().f_back.f_lineno

def b2hex(data):
	if PY2:
		return ' '.join(map(lambda b: format(ord(b), "02x"), data))
	else:
		if type(data) is str:
			return ' '.join(map(lambda b: format(ord(b), "02x"), data))
		else:
			return ' '.join(map(lambda b: format(b, "02x"), data))

def hexdump(pk, prefix='', file=None):
	if file is not None:
		line_fmt = u'{0} {2}{3}{4}'
		hex_width = 4
		blk_width = 4
	else:
		line_fmt = u'{0}{1:8x}  {2}{3} {4}'
		hex_width = HEX_WIDTH
		blk_width = 8

	lpk = len(pk)
	ofs = 0
	hexofs = 0
	hexlen = 0
	hexstr = ''
	ascstr = ''
	ascstr_width = int(hex_width * 100 / 32.0 - 0.5)  # 32h = 100a, 16h = 50a
	while ofs < lpk:
		hexstr += b2hex(pk[ofs:ofs+blk_width])
		hexstr += ' '
		if PY2:
			ascstr += ''.join(map(lambda b: b if ord(b) >= 0x20 and ord(b) < 0x7f else '.', pk[ofs:ofs+blk_width]))
		else:
			ascstr += ''.join(map(lambda b: chr(b) if b >= 0x20 and b < 0x7f else '.', pk[ofs:ofs+blk_width]))
		hexlen += blk_width
		ofs += blk_width
		
		if hexlen >= hex_width or ofs >= lpk:
			txt = line_fmt.format(prefix, hexofs, hexstr,
				u' '*(ascstr_width-len(hexstr)), ascstr)
			
			if file is not None:
				file.write((txt + u'\n').encode('utf-8'))
			else:
				print(txt)

			hexofs = ofs
			hexstr = ''
			hexlen = 0
			ascstr = ''
		else:
			hexstr += ' '
			ascstr += ' '

def termsize():
	import os
	env = os.environ
	def ioctl_GWINSZ(fd):
		try:
			import fcntl, termios, struct, os
			cr = struct.unpack('hh', fcntl.ioctl(fd, termios.TIOCGWINSZ, '1234'))
		except:
			return
		return cr
	cr = ioctl_GWINSZ(0) or ioctl_GWINSZ(1) or ioctl_GWINSZ(2)
	if not cr:
		try:
			fd = os.open(os.ctermid(), os.O_RDONLY)
			cr = ioctl_GWINSZ(fd)
			os.close(fd)
		except:
			pass
	if not cr:
		try:
			cr = (env['LINES'], env['COLUMNS'])
		except:
			cr = (25, 80)
	return int(cr[1]), int(cr[0])

sz = termsize()
sw = sz[0]
sh = sz[1]

def terrible(exctype, value, traceback):
	print(exctype)
	print(value)
	print(traceback)
	os._exit(0)

#sys.excepthook = terrible

## END DEBUG HELPERS
##


class Unrecv(object):
	def __init__(self, s):
		self.s = s
		self.buf = b''
	
	def recv(self, nbytes):
		if self.buf:
			ret = self.buf[:nbytes]
			self.buf = self.buf[nbytes:]
			return ret
		
		try:
			return self.s.recv(nbytes)
		except:
			return b''
	
	def unrecv(self, buf):
		self.buf = buf + self.buf


class UnbufferedHttpd(object):
	def __init__(self, hwm, addr, port):
		srv = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
		srv.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
		srv.bind((addr, port))
		srv.listen(max_threads)
		
		while True:
			try:
				sck = srv.accept()
			except:
				break
			
			hwm.accept(sck)


class UnbufferedClient(object):
	def __init__(self, log_fwd, sck, addr):
		self.log_fwd = log_fwd
		self.s = sck
		self.sr = Unrecv(sck)
		self.addr = addr
		self.ok = True
		
		self.t0 = time.time()
		self.bytes_now = 0
		self.bytes_eos = 0
		
		self.workload = 500
		self.next_workload = 0
		self.workload_mutex = threading.Lock()
		
		thr = threading.Thread(target=self.workload_thr)
		thr.daemon = True
		thr.start()
	
	def log(self, msg):
		self.log_fwd(self.addr, msg)
	
	def workload_thr(self):
		while self.s:
			time.sleep(0.5)
			with self.workload_mutex:
				self.workload = (
					0.7 * self.workload +
					0.3 * self.next_workload)
				self.next_workload = 0
	
	def run(self):
		headerlines = self.read_header()
		if not self.ok:
			return
		
		self.headers = {}
		mode, self.req, _ = headerlines[0].split(' ')
		
		for header_line in headerlines[1:]:
			k, v = header_line.split(':', 1)
			self.headers[k.lower()] = v.strip()
		
		try:
			self.bytes_eos = int(self.headers['content-length'])
		except:
			pass
		
		if mode == 'GET':
			self.handle_get()
		elif mode == 'POST':
			self.handle_post()
		else:
			self.loud_reply(u'invalid HTTP mode "{0}"'.format(mode))
	
	def panic(self, msg):
		self.log('ERROR: client disconnected ({0})'.format(msg).upper())
		self.ok = False
		self.s.close()
	
	def read_header(self):
		ret = b''
		while True:
			if ret.endswith(b'\r\n\r\n'):
				break
			elif ret.endswith(b'\r\n\r'):
				n = 1
			elif ret.endswith(b'\r\n'):
				n = 2
			elif ret.endswith(b'\r'):
				n = 3
			else:
				n = 4
			
			buf = self.sr.recv(n)
			if not buf:
				self.panic('headers')
				break
			
			ret += buf
		
		return ret[:-4].decode('utf-8', 'replace').split('\r\n')
	
	def reply(self, body):
		header = 'HTTP/1.1 200 OK\r\nContent-Type: text/html\r\nContent-Length: {0}\r\nConnection: close\r\n\r\n'.format(len(body)).encode('utf-8')
		if self.ok:
			self.s.send(header + body)
		
		self.s.close()
		return body
	
	def loud_reply(self, body):
		self.log(body)
		self.reply(b'<pre>' + body.encode('utf-8'))
	
	def handle_get(self):
		self.log('GET  {0} {1}'.format(self.addr[0], self.req))
		self.reply(b'<form method="post" enctype="multipart/form-data"><h5><input type="file" name="f" multiple><h5><input type="submit" value="start upload">')
	
	def handle_post(self):
		self.log('POST {0} {1}'.format(self.addr[0], self.req))
		
		try:
			if self.headers['expect'].lower() == '100-continue':
				self.s.send(b'HTTP/1.1 100 Continue\r\n\r\n')
		except:
			pass
		
		form_segm = self.read_header()
		if not self.ok:
			return
		
		self.boundary = b'\r\n' + form_segm[0].encode('utf-8')
		for ln in form_segm[1:]:
			self.log(ln)
		
		fn0 = 'inc.{0:.9f}'.format(time.time())
		if nullwrite:
			fn0 = '/dev/null'

		files = []
		t0 = time.time()
		for nfile in range(99):
			fn = fn0
			if not nullwrite:
				fn = '{0}.{1}'.format(fn0, nfile)
			
			with open(fn, 'wb') as f:
				self.log('writing to {0}'.format(fn))
				sz, sha512 = self.handle_multipart(f)
				if sz == 0:
					break
				
				files.append([sz, sha512])
			
			buf = self.sr.recv(2)
			
			if buf == b'--':
				# end of multipart
				break
			
			if buf != b'\r\n':
				return self.loud_reply(u'protocol error')
			
			header = self.read_header()
			if not self.ok:
				break
			
			form_segm += header
			for ln in header:
				self.log(ln)
		
		td = time.time() - t0
		sz_total = sum(x[0] for x in files)
		spd = (sz_total / td) / (1024*1024)
		
		status = 'OK'
		if not self.ok:
			status = 'ERROR'
		
		msg = u'{0} // {1} bytes // {2:.3f} MiB/s\n'.format(status, sz_total, spd)
		
		for sz, sha512 in files:
			msg += u'sha512: {0} // {1} bytes\n'.format(sha512[:56], sz)
			# truncated SHA-512 prevents length extension attacks;
			# using SHA-512/224, optionally SHA-512/256 = :64
		
		self.loud_reply(msg)

		if not nullwrite:
			with open(fn0 + '.txt', 'wb') as f:
				f.write((u'\n'.join(unicode(x) for x in [
					u':'.join(unicode(x) for x in self.addr),
					u'\n'.join(form_segm),
					msg.rstrip()
					]) + '\n').encode('utf-8'))
	
	def handle_multipart(self, ofd):
		tlen = 0
		hashobj = hashlib.sha512()
		for buf in self.extract_filedata():
			tlen += len(buf)
			hashobj.update(buf)
			ofd.write(buf)
		
		return tlen, hashobj.hexdigest()
	
	def extract_filedata(self):
		blen = len(self.boundary)
		while True:
			buf = self.sr.recv(32768)
			if not buf:
				# abort: client disconnected
				self.panic('outer')
				return
			
			with self.workload_mutex:
				self.next_workload += 1
				self.bytes_now += len(buf)
			
			#hexdump(buf, _lno())
			
			while True:
				ofs = buf.find(self.boundary)
				if ofs != -1:
					self.sr.unrecv(buf[ofs + blen:])
					#hexdump(buf[:ofs], 'ret')
					#hexdump(self.sr.buf, 'rem')
					yield buf[:ofs]
					return
				
				d = len(buf) - blen
				if d > 0:
					# buffer growing large; yield everything except
					# the part at the end (maybe start of boundary)
					yield buf[:d]
					buf = buf[d:]
				
				# look for boundary near the end of the buffer
				for n in range(1, len(buf)+1):
					if not buf[-n:] in self.boundary:
						n -= 1
						break
				
				if n == 0 \
				or not self.boundary.startswith(buf[-n:]):
					# no boundary contents near the buffer edge
					break
				
				if blen == n:
					# EOF: found boundary
					yield buf[:-n]
					return
				
				buf2 = self.sr.recv(32768)
				if not buf2:
					# abort: client disconnected
					self.panic('inner')
					return
				
				with self.workload_mutex:
					self.next_workload += 1
					self.bytes_now += len(buf2)
				
				buf += buf2
			
			yield buf


class HttpWorkerProcess(object):
	"""
	one subprocess with several threads,
	each processing one accepted http client
	"""
	def __init__(self, nid, q_pend, q_yield):
		self.nid = nid
		self.q_pend = q_pend
		self.q_yield = q_yield
		self.mutex = threading.Lock()

		self.clients = []
		self.last_workload = -1
		
		signal.signal(signal.SIGINT, signal_nuller)
	
		thr = threading.Thread(target=self.status_thread)
		thr.daemon = True
		thr.start()
		
		while True:
			while True:
				# TODO: replace with semaphors
				time.sleep(0.01)
				with self.mutex:
					if len(self.clients) < max_threads:
						break
			
			self.log('awaiting work')
			d = q_pend.get()
			
			self.log('\033[32mcmd [' + d[0] + ']')
			if d[0] == 'terminate':
				self.log('bye')
				sys.exit(0)
				return
			
			# manually unpickle socket to support <py3.4
			sck = pickle.loads(d[1])
			cli = UnbufferedClient(self.clog, *sck)
			self.clients.append(cli)
			
			thr = threading.Thread(target=self.worker_thread, args=(cli,))
			thr.daemon = True
			thr.start()
	
	def worker_thread(self, cli):
		"""one client thread, processing one http client"""
		with self.mutex:
			self.q_yield.put(['threads', len(self.clients)])
		
		cli.thread = threading.current_thread()
		cli.run()
		
		with self.mutex:
			self.clients.remove(cli)
			self.q_yield.put(['dropclient', str(cli.addr)])
			self.q_yield.put(['threads', len(self.clients)])
			self.emit_workload()
		
		self.log('client-handler terminated')

	def status_thread(self):
		"""
		single background thread per subprocess (self);
		periodically emits workload info to the HttpWorkerManager
		"""
		while True:
			time.sleep(0.1)
			with self.mutex:
				self.emit_workload()
				for cli in self.clients:
					self.q_yield.put(['progress', str(cli.addr), [cli.bytes_now, cli.bytes_eos]])

	def emit_workload(self):
		workload = 0
		for cli in self.clients:
			with cli.workload_mutex:
				workload += cli.workload
		
		if workload != self.last_workload:
			self.q_yield.put(['workload', workload])
	
	def log(self, msg):
		self.q_yield.put(['log', '<proc{0}>'.format(self.nid), msg])

	def clog(self, who, msg):
		self.q_yield.put(['log', who, msg])


class HttpWorkerManager(object):
	"""
	the top-level object which manages HttpWorkerProcess instances
	(each HWP is a subprocess which in turn has several threads)
	"""
	def __init__(self, gui, num_cores):
		self.gui = gui
		self.procs = []
		self.mutex = threading.Lock()
		
		for n in range(num_cores):
			q_pend = mp.Queue(1)
			q_yield = mp.Queue(64)
			
			proc = mp.Process(target=HttpWorkerProcess, args=(n, q_pend, q_yield))
			proc.nid = n
			proc.threads = 0
			proc.workload = 0
			proc.clients = {}
			proc.q_pend = q_pend
			proc.q_yield = q_yield
			
			thr = threading.Thread(target=self.yieldcollector, args=(proc,))
			thr.daemon = True
			thr.start()
			
			self.procs.append(proc)
			proc.start()
	
	def yieldcollector(self, proc):
		while True:
			obj = proc.q_yield.get()
			with self.mutex:
				if obj[0] == 'threads':
					proc.threads = obj[1]
				
				elif obj[0] == 'workload':
					proc.workload = obj[1]
				
				elif obj[0] == 'dropclient':
					try:
						del proc.clients[obj[1]]
					except:
						pass
				
				elif obj[0] == 'progress':
					proc.clients[obj[1]] = obj[2]
				
				elif obj[0] == 'log':
					self.gui.log(u'{0:<27} \033[0m{1}'.format(str(obj[1]), obj[2]))
	
	def accept(self, sck):
		while True:
			# TODO: replace with semaphors
			for proc in sorted(self.procs, key=lambda x: x.workload):
				if proc.threads < max_threads:
					# can't put unpickled sockets <3.4
					buf = MemesIO()
					ForkingPickler(buf).dump(sck)
					proc.q_pend.put(['socket', buf.getvalue()])
					proc.workload += 500
					return
			
			time.sleep(0.01)
	
	def terminate(self):
		for proc in self.procs:
			self.gui.log('<@>\033[31m sendmsg terminate proc{0}'.format(proc.nid))
			proc.q_pend.put(['terminate'])
		
		self.terminated = True


def signal_nuller(signal, frame):
	print('ok.jpg')

	
sig_ctr = 0
def signal_handler(signal, frame):
	global sig_ctr
	# do not invoke complex logic here
	# (push subprocess queues, with(mutex), etc)
	# as almost anything can deadlock from within a sigtrap
	sig_ctr += 1
	if sig_ctr >= 3:
		os._exit(0)


def signal_processor(gui, *workers):
	while True:
		time.sleep(0.1)
		if sig_ctr == 1:
			gui.log('<@>\033[31m shutting down')
			for w in workers:
				w.terminate()
			
			ok = False
			while not ok:
				ok = True
				time.sleep(0.1)
				for w in workers:
					if not hasattr(w, 'terminated'):
						ok = False
						break
			
			gui.log('<@>\033[31m bye')
			sys.exit(0)
			#os._exit(0)  # TODO: sys.exit gets stuck


class GUI(object):
	def __init__(self):
		self.hwm = None
		self.hist = []
		self.mutex = threading.Lock()
		
		thr = threading.Thread(target=self.worker)
		thr.daemon = True
		thr.start()
	
	def log(self, msg):
		ts = datetime.utcnow().strftime('%Y-%m-%d, %H:%M:%S')
		with self.mutex:
			msg = '\033[36m{0}  \033[33m{1}'.format(ts, msg)
			for x in msg.split('\n'):
				self.hist.append(x)
			
			if len(self.hist) > 400:
				self.hist = self.hist[200:]
	
	def worker(self):
		#return
		while True:
			m = u''
			if self.hwm:
				with self.hwm.mutex:
					for proc in self.hwm.procs:
						m += '\033[0mproc{0}: clients({1}) load({2})\n\033[32m'.format(proc.nid, proc.threads, proc.workload)
						for cli, (pos, eos) in proc.clients.items():
							perc = 0
							if eos > 0 and pos < eos:
								perc = 100 * (pos * 1.0 / eos)
							
							m += '  {0}: {1} total, {2} now, {3:.2f}%\n'.format(
								cli, eos, eos, perc)
			
			m += '\033[35m'
			left = sh - (m.count('\n') + 2)
			with self.mutex:
				for msg in self.hist[-left:]:
					m += '\n' + msg
			
			print('\033[H' + m.replace('\n', '\033[K\n'), end='\033[J\033[0m\n')
			time.sleep(0.2)


if __name__ == '__main__':
	if len(sys.argv) > 1:
		bind_port = int(sys.argv[1])

	signal.signal(signal.SIGINT, signal_handler)

	ip = '127.0.0.1'
	if bind_ip != ip:
		s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
		try:
			s.connect(('10.255.255.255', 1))
			ip = s.getsockname()[0]
		except:
			pass
		s.close()

	if cores == 0:
		cores = mp.cpu_count()

	gui = GUI()
	hwm = HttpWorkerManager(gui, cores)
	gui.hwm = hwm  # this-is-fine.jpg
	
	thr = threading.Thread(target=signal_processor, args=(gui,hwm))
	thr.daemon = True
	thr.start()
	
	gui.log('<@> {0} threads @ http://{1}:{2}/'.format(cores, ip, bind_port))
	UnbufferedHttpd(hwm, bind_ip, bind_port)
	thr.join()
